#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <stdarg.h>
#include <errno.h>
#include "send_cmd.h"
#include "daemon.h"
#include "tiltmeter.h"
#include "control_interface.h"
#include "cmdline.h"

typedef struct {
	int connection;
	int out_buffer_pending;
	int out_buffer_pending_start;
	int out_buffer_free[2];
	int out_buffer_size[2];
	char *out_buffer[2];
	
	int in_buffer_start;
	int in_buffer_free;
	int in_buffer_size;
	char *in_buffer;
	} CONTROL_CONNECTION;

extern struct gengetopt_args_info args_info;
	
CONTROL_CONNECTION *control_connections;
int cc_free=0, cc_size=0;

int listen_connection=-1;

void parse_control_buffer(CONTROL_CONNECTION *cc);

void init_cc(void)
{
struct sockaddr_in sa;

cc_size=10;
control_connections=do_alloc(cc_size, sizeof(*control_connections));

/* Perhaps add option to listen on TCP connections later ? */
listen_connection=socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
if(listen_connection<0) {
	perror("Cannot create control connection:");
	exit(-1);
	}

memset(&sa, 0, sizeof(sa));
sa.sin_family=AF_INET;
sa.sin_port=0xda8a;
sa.sin_addr.s_addr=INADDR_ANY;

if(bind(listen_connection, &sa, sizeof(sa))<0) {
	perror("Cannot bind control connection:");
	exit(-1);
	}

if(listen(listen_connection, 10)<0) {
	perror("Cannot listen on control connection:");
	exit(-1);
	}
}

void accept_connection(void)
{
int i;
CONTROL_CONNECTION *cc;
struct sockaddr addr;
socklen_t socklen;
int con;

con=accept(listen_connection, &addr, &socklen);
if(con<0) {
	perror("accept_connection:");
	return;
	}

for(i=0;i<cc_free && control_connections[i].connection>=0;i++);

if(i>=cc_size) {
	fprintf(stderr, "**** ERROR: connection limit reached\n");
	close(con);
	return;
	}
	
	
cc=&(control_connections[i]);

cc->connection=con;
cc->out_buffer_free[0]=0;
cc->out_buffer_free[1]=0;
cc->in_buffer_start=0;
cc->in_buffer_free=0;
cc->out_buffer_pending=0;
cc->out_buffer_pending_start=0;

if(i==cc_free) {
	cc_free++;
	cc->out_buffer_size[0]=1024*1024;
	cc->out_buffer_size[1]=1024*1024;
	cc->out_buffer[0]=do_alloc(cc->out_buffer_size[0], 1);
	cc->out_buffer[1]=do_alloc(cc->out_buffer_size[1], 1);

	cc->in_buffer_size=1024*1024;	
	cc->in_buffer=do_alloc(cc->in_buffer_size, 1);
	}
if(args_info.tiltmeter_mode_arg)send_header();
fprintf(stderr, "Accepted connection %d\n", i);
}
	
	
int max_cc_fd(void)
{
int i, con;
con=listen_connection;
for(i=0;i<cc_free;i++)
	if(con<control_connections[i].connection)con=control_connections[i].connection;
return(con);
}

void set_cc_fds(fd_set *fd_read, fd_set *fd_write)
{
int i;
FD_SET(listen_connection, fd_read);
for(i=0;i<cc_free;i++)
	if(control_connections[i].connection>=0) {
		if(control_connections[i].out_buffer_pending_start<control_connections[i].out_buffer_free[control_connections[i].out_buffer_pending])FD_SET(control_connections[i].connection, fd_write);
		FD_SET(control_connections[i].connection, fd_read);
		}
}

void write_cc_data(fd_set *fd_write)
{
int i;
int written;
CONTROL_CONNECTION *cc;
for(i=0;i<cc_free;i++)
	if(control_connections[i].connection>=0 && FD_ISSET(control_connections[i].connection, fd_write)) {
		cc=&(control_connections[i]);
		written=write(cc->connection, &(cc->out_buffer[cc->out_buffer_pending][cc->out_buffer_pending_start]), cc->out_buffer_free[cc->out_buffer_pending]-cc->out_buffer_pending_start);
		if(written>0) {
			cc->out_buffer_pending_start+=written;
			if(cc->out_buffer_pending_start>=cc->out_buffer_free[cc->out_buffer_pending]) {
				cc->out_buffer_free[cc->out_buffer_pending]=0;
				cc->out_buffer_pending_start=0;
				cc->out_buffer_pending=!cc->out_buffer_pending;
				}
			}
		}
}

void read_cc_data(fd_set *fd_read)
{
int i;
int bytes_read;
CONTROL_CONNECTION *cc;
for(i=0;i<cc_free;i++)
	if(control_connections[i].connection>=0 && FD_ISSET(control_connections[i].connection, fd_read)) {
		cc=&(control_connections[i]);
		bytes_read=read(cc->connection, &(cc->in_buffer[cc->in_buffer_free]), cc->in_buffer_size-cc->in_buffer_free);
		if(bytes_read<0) perror("read");
		if(bytes_read==0 || (bytes_read<0 && bytes_read!=EAGAIN)) {
			close(cc->connection);
			cc->connection=-1;
			fprintf(stderr, "Closed connection %d\n", i);
			continue;
			}
		if(bytes_read>0)cc->in_buffer_free+=bytes_read;
		parse_control_buffer(cc);
		}

if(FD_ISSET(listen_connection, fd_read))accept_connection();
}


void parse_control_buffer(CONTROL_CONNECTION *cc)
{
int i;	

while(cc->in_buffer_start<cc->in_buffer_free) {
	
	for(i=cc->in_buffer_start;i<cc->in_buffer_free && cc->in_buffer[i]!='\n';i++);
	
	if(cc->in_buffer[i]=='\n') {
		cc->in_buffer[i]=0;
		parse_control_cmd(&(cc->in_buffer[cc->in_buffer_start]), i-cc->in_buffer_start);
		cc->in_buffer_start=i+1;
		
		if(cc->in_buffer_start>=cc->in_buffer_free) {
			cc->in_buffer_free=0;
			cc->in_buffer_start=0;
			}
		} else {
		/* This is the last remaining chunk, move it to the start*/
		if(cc->in_buffer_start>0) {
			memmove(cc->in_buffer, &(cc->in_buffer[cc->in_buffer_start]), cc->in_buffer_free-cc->in_buffer_start);
			cc->in_buffer_free-=cc->in_buffer_start;
			cc->in_buffer_start=0;
			}
		/* we cannot parse more lines, exit loop */
		break;
		}	
	}	
}


void parse_control_cmd(char *str, int count)
{
if(!strncmp(str, "dac_cmd ", 8)) {
	int port, reg, value;
	
	sscanf(str, "dac_cmd %d %d %d", &port, &reg, &value);
	send_dac_cmd(port, reg, value);
	} else
if(!strncmp(str, "set_dac ", 8)) {
	int port, channel;
	double value;
	
	sscanf(str, "set_dac %d %d %lf", &port, &channel, &value);
	set_dac(port, channel, value);
	} else
if(!strncmp(str, "set_dac4x ", 10)) {
	int port;
	double ch0, ch1, ch2, ch3;
	
	sscanf(str, "set_dac4x %d %lf %lf %lf %lf", &port, &ch0, &ch1, &ch2, &ch3);
//	fprintf(stderr, "set_dac4x %d %lf %lf %lf %lf\n", port, ch0, ch1, ch2, ch3);
	set_dac4x(port, ch0, ch1, ch2, ch3);
	} else
if(!strncmp(str, "set_variable ", 13)) {
        set_variable(str+13);
        } else
if(!strncmp(str, "return_vars", 11)) {
        broadcast_all_variables();
        } else
if(!strncmp(str, "set_leds ", 9)) {
	int x;
	
	sscanf(str, "set_leds %d", &x);
	set_leds(x);
	} else
if(!strncmp(str, "get_status", 10)) {
	request_status();
	} else
if(!strncmp(str, "reinitialize", 12)) {
	init_board1();
	sleep(3);
	start_board1();	
	sleep(3);
	start_board2();	
	} else
if(!strncmp(str, "stop_board", 10)) {
	stop_board();	
	} else
	{
	fprintf(stderr, "*** UNRECOGNIZED COMMAND: %s\n", str);
	}
}

static char tmp_s[1000000];

void broadcast(const char *fmt, ...)
{
va_list ap;
int n;
int i;
CONTROL_CONNECTION *cc;

va_start(ap, fmt);
n=vsnprintf(tmp_s, 1000000-1, fmt, ap);
va_end(ap);

if(n<=0)return;

for(i=0;i<cc_free;i++) {
	if(control_connections[i].connection<0)continue;
	cc=&(control_connections[i]);
	
	if(cc->out_buffer_free[cc->out_buffer_pending]+n-1<cc->out_buffer_size[cc->out_buffer_pending]) {
		memcpy(&(cc->out_buffer[cc->out_buffer_pending][cc->out_buffer_free[cc->out_buffer_pending]]), tmp_s, n);
		cc->out_buffer_free[cc->out_buffer_pending]+=n;
		//fprintf(stderr, "%d ", cc->out_buffer_free[cc->out_buffer_pending]);

		}
	}
}

static VARIABLE *vars=NULL;
static int vars_free=0; 
static int vars_size=0;

void add_variables(VARIABLE *v, int count)
{
VARIABLE *p;

if(count<0)
        for(count=0;v[count].type>0;count++);

if(vars_free+count>vars_size) {
	vars_size=2*vars_size+count;
	p=do_alloc(vars_size, sizeof(*vars));
	if(vars_free>0)memcpy(p, vars, vars_free*sizeof(*vars));
	free(vars);
	vars=p;
	}
memcpy(&(vars[vars_free]), v, count*sizeof(*vars));
vars_free+=count;
}

void broadcast_variables(VARIABLE *v, int count)
{
int i;
char *s;
int s_len;
s_len=200;

if(count<0)
	for(count=0;v[count].type>0;count++);

for(i=0;i<count;i++)s_len+=strlen(v[i].format)+strlen(v[i].name)+20;
s=do_alloc(s_len, sizeof(*s));

sprintf(s, "SET_VARIABLE");
for(i=0;i<count;i++) {
	switch(v[i].type) {
		case TYPE_INT:
			sprintf(s+strlen(s), " %s ", v[i].name);
			sprintf(s+strlen(s), v[i].format, *(int *)v[i].var);
			break;
                case TYPE_DOUBLE:
                        sprintf(s+strlen(s), " %s ", v[i].name);
                        sprintf(s+strlen(s), v[i].format, *(double *)v[i].var);
                        break;

		}
	}
broadcast("%s\n", s);
free(s);
}

void broadcast_all_variables(void)
{
broadcast_variables(vars, vars_free);
}

void set_variable(char *s)
{
int i;
int l=strlen(s);
int k;
double a;

fprintf(stderr, "%s\n", s);

for(i=0;i<vars_free;i++) {
	k=strlen(vars[i].name);
	if(k+1>l)continue;
	if(s[k]!=' ')continue;
	if(strncmp(s, vars[i].name, k))continue;

	sscanf(s+k+1, vars[i].scan_format, vars[i].var);
	
	if(vars[i].callback!=NULL)vars[i].callback(vars[i].priv);
	return;
	}
}
