#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netpacket/packet.h>
#include <net/ethernet.h> /* the L2 protocols */
#define _GNU_SOURCE
#define __USE_GNU
#include <sched.h>
#include "daemon.h"
#include "receive_packet.h"
#include "send_cmd.h"
#include "control_interface.h"
#include "tiltmeter.h"
#include "cmdline.h"

int raw_sock;

unsigned char buffer[65536];

struct gengetopt_args_info args_info;



typedef struct {
	unsigned char dst_mac[6];
	unsigned char src_mac[6];
	unsigned char type[2];
	unsigned char packet_count[6];
	} ETHERNET_DATA_HEADER;

typedef struct {
	unsigned char type[2];
	unsigned char sample_count[6];
	unsigned char channel_data[12];
	} SUBPACKET_HEADER;
	
	
typedef struct {
	unsigned char type[2];
	unsigned char switches[2];
	unsigned tx_counter;
	u64 general_config;
	u64 port_config;
	unsigned char source_mac[6];
	unsigned char filler1[2];
	unsigned char destination_mac[6];
	unsigned char filler2[2];
	unsigned sclk_counter;	
	} STATUS_SUBPACKET;

FILE *data_log;

void *do_alloc(long a, long b)
{
void *r;
int i=0;
if(a<1)a=1;
if(b<1)b=1;
r=calloc(a,b);
while(r==NULL){
	fprintf(stderr,"Could not allocate %ld chunks of %ld bytes each (%ld bytes total), current memory usage %ld\n",a,b,a*b, MEMUSAGE);
	if(i>10)exit(-1);
	sleep(10);
	r=calloc(a,b);
	i++;
	}
return r;
}

double get_time(void)
{
struct timeval tv;
gettimeofday(&tv, NULL);
return(tv.tv_sec*1.0+tv.tv_usec*1e-6);
}
	
int main(int argc, char *argv []) 
{
int len, last_len;
int i;
ETHERNET_HEADER *header=buffer;
fd_set fd_read, fd_write;
int data_log_fd, max_fd;
int empty_buffer=1;
cpu_set_t cs;
int cpu, node;

if(sizeof(u64)!=8) {
	fprintf(stderr, "*** COMPILATION ERROR: u64 is not 8 bytes long\n");
	exit(-1);
	}

if(cmdline_parser(argc, argv, &args_info))exit(-1);

for(i=0;i<args_info.config_given;i++)
        if(cmdline_parser_configfile(args_info.config_arg[i], &args_info, 0, 0, 0))exit(-1);

if(!args_info.board_mac_given) {
	fprintf(stderr, "You must specify data acquisition board MAC address (--board-mac)\n");
	exit(-1);
	}

if(!args_info.daemon_mac_given) {
	fprintf(stderr, "You must specify data host computer MAC address (--daemon-mac)\n");
	exit(-1);
	}

if(args_info.tiltmeter_mode_arg) {
	init_tiltmeter();
	}

if(sched_getaffinity(0, sizeof(cs), &cs)<0) {
	fprintf(stderr, "Failed to obtain cpu affinity mask\n");
	}

for(cpu=0;!CPU_ISSET(cpu, &cs) && cpu<64;cpu++);

fprintf(stderr, "Locking to cpu %d\n", cpu);

CPU_ZERO(&cs);
CPU_SET(cpu, &cs);

if(sched_setaffinity(0, sizeof(cs), &cs)<0) {
	fprintf(stderr, "Failed to set cpu affinity\n");
	}

raw_sock=socket(AF_PACKET, SOCK_RAW, 0x8ada);

if(raw_sock<0) {
	perror("Error creating raw socket");
	exit(-1);
	}

if(args_info.data_log_given) {
	data_log=fopen(args_info.data_log_arg, "w");
	if(data_log==NULL) {
		perror("Cannot open data_log file");
		exit(-1);
		}
		
	setbuffer(data_log, do_alloc(10*1000*1000, 1), 10*1000*1000);


	data_log_fd=fileno(data_log);
	} else {
	data_log=NULL;
	data_log_fd=-1;
	}


init_cc();

init_out_buffer();
init_board1();
sleep(3);
start_board1();
sleep(3);
start_board2();

while(1) {
	FD_ZERO(&fd_read);
	FD_ZERO(&fd_write);

	if(!empty_buffer && data_log_fd>=0)FD_SET(data_log_fd, &fd_write);
	FD_SET(raw_sock, &fd_read);

	set_cc_fds(&fd_read, &fd_write);
	
	max_fd=max_cc_fd();
	if(data_log_fd>max_fd)max_fd=data_log_fd;
	if(raw_sock>max_fd)max_fd=raw_sock;
	
	select(max_fd+1, &fd_read, &fd_write, NULL, NULL);

	if(FD_ISSET(raw_sock, &fd_read)) {
		empty_buffer=0;
		
		len=recv(raw_sock, buffer, 65536, 0);
		if(len<0) {
			perror("Error receiving message");
			continue;
			}
		parse_packet(buffer, len);

		last_len=len;
		while((len=recv(raw_sock, buffer, 65536, MSG_DONTWAIT))>0) {
			parse_packet(buffer, len);
			}
		}
	
	//fprintf(stderr, "len=%d\n", last_len);
	if(data_log_fd>=0 && FD_ISSET(data_log_fd, &fd_write)) {
		empty_buffer=1;
		fflush(data_log);
		}
	
	read_cc_data(&fd_read);
	write_cc_data(&fd_write);	
	}

return 0;
}
