#!/usr/bin/env wish

foreach {var value} {
	line_count 0
	dac0_value 0.0
	dac1_value 0.0
	dac2_value 0.0
	dac3_value 0.0
	control_mode 0
	mode_switch 1
	mark ""
	header {x}
	} { 
	global $var
	set $var $value
	}

global CONTROL
set CONTROL [socket "localhost" 35546]

proc wrap_up {} {
global CONTROL
close $CONTROL
exit 0
}

wm protocol . WM_DELETE_WINDOW wrap_up

global LOG
set LOG [open "capture.log" "w"]

proc read_data { c } {
global LOG

fconfigure $c -blocking true
gets $c line
fconfigure $c -blocking false

if { $line == "" } { return }

switch -exact -- [lindex $line 0] {
	PHASE   {
		set_phase_row .phases [lindex $line 1] [lrange $line 2 end]
		}

	LVDT_V {
		add_lvdt_data .g1 [lrange $line 2 10]

		global header control_mode position mark

		foreach var $header value $line {
			global $var
			set $var $value
			#puts "$var=$value"
			}

		catch { 
			set position [expr $lvdt2_coarse - $lvdt1_coarse]
			}
		puts $LOG "$line $control_mode \"$mark\""
		}

	LVDT_HEADER {
		global header
		set header $line
		puts $LOG "$line control_mode mark"
		}

	SET_VARIABLE {
		foreach {var value} [lrange $line 1 end] {
			global $var
			set $var $value
			}
		}

	MODE_SWITCH {
		global mode_switch
		set mode_switch 1
		}

	default {
		puts stderr $line
		puts $LOG $line
		#.line delete 0 end
		#.line insert 0 $line
		}
	}

}

proc send_cmd { s } {
global CONTROL
fconfigure $CONTROL -blocking true
puts $CONTROL $s
flush $CONTROL
fconfigure $CONTROL -blocking false
}

proc make_phase_widget { w } {

frame $w

for { set i 0 } { $i < 8 } { incr i } {
	label $w.l$i -text "ch$i"
	grid $w.l$i -row [expr $i+1] -column 0 -sticky news
	}

for { set j 0 } { $j < 8 } { incr j } {
	label $w.lp$j -text "Phase $j"
	grid $w.lp$j -column [expr $j+1] -row 0 -sticky news
	}

for { set i 0 } { $i < 8 } { incr i } {
	for { set j 0 } { $j < 8 } { incr j } {
		entry $w.phase_${i}_$j
		grid $w.phase_${i}_$j -row [expr $i+1] -column [expr $j+1] -sticky news
		}
	}
}

proc set_phase_row { w row data } {
for { set i 0 } { $i < 8 } { incr i } {
	$w.phase_${row}_$i delete 0 end
	$w.phase_${row}_$i insert 0 [lindex $data [expr $i]]
	}
}

proc make_graph_widget { w {width 300} { height 200} {max_points 1000} {incr_points 20} } {
frame $w
canvas $w.c -width $width -height $height -background white
grid $w.c

global $w.data $w.settings $w.updating
set $w.data {}
set $w.updating 0

foreach var {max_points incr_points} {
	set $w.settings($var) [set $var]
	}

$w.c create text [expr $width/2] [expr $height/2] -text "No data" -fill red -tag "no_data"
}

proc add_graph_point { w x y } {
global $w.data $w.settings $w.updating

lappend $w.data $x $y

foreach var { max_points incr_points } {
	set $var [set $w.settings($var)]
	}

if { [llength [set $w.data]] > $max_points+$incr_points } { 
	set $w.data [lrange [set $w.data] $incr_points end]

	if { ! [set $w.updating] } {
		set $w.updating 1
		after idle "update_graph $w"
		}
	}

if { [llength [set $w.data]] <= $max_points && ([llength [set $w.data]] % $incr_points==0) } {
	if { ! [set $w.updating] } {
		set $w.updating 1
		after idle "update_graph $w"
		}
	}

}


proc update_graph { w } {
global $w.data $w.settings $w.updating $w.updating

if { [llength [set $w.data]] >= 4 } {
	set x0 [lindex [set $w.data] 0]
	set y0 [lindex [set $w.data] 1]

	set max_x -1e20
	set min_x 1e20
	set max_y -1e20
	set min_y 1e20
	set sum_y 0.0
	set sum_yy 0.0
	set count_y 0

	foreach {x y } [set $w.data] {
		if { $y== "Inf" || $y == "-Inf" || "$y" == "nan" || "$y" == "-nan" || $y == "inf" || $y== "-inf"} { continue }
		if { $x > $max_x } { set max_x $x }
		if { $x < $min_x } { set min_x $x }
		if { $y > $max_y } { set max_y $y }
		if { $y < $min_y } { set min_y $y }

		if { [catch {
		set sum_y [expr $sum_y+$y]
		set sum_yy [expr $sum_yy+$y*$y]
			} ]} { puts "$sum_y $sum_yy $y" }
		incr count_y
		}

	set width [$w.c cget -width]
	set height [$w.c cget -height]
	set inv_x 0.0
	set inv_y 0.0
	catch {
		set inv_x [expr $width/($max_x-$min_x+1e-56)]
		set inv_y [expr $height/($max_y-$min_y+1e-56)]
		}


	if { $max_x < $min_x || $max_y < $min_y } { 
		set $w.updating 0
		return 
		}

	$w.c delete no_data
	$w.c delete curve

	foreach {x y } [lrange [set $w.data] 2 end] {
		if { $y== "Inf" || $y == "-Inf" } { continue }
		catch {
			$w.c create line [expr ($x0-$min_x)*$inv_x] [expr ($max_y-$y0)*$inv_y] [expr ($x-$min_x)*$inv_x] [expr ($max_y-$y)*$inv_y] -fill blue -tag curve
			}
		set x0 $x
		set y0 $y
		}

	$w.c delete info
	$w.c create text [expr $width-120] 20 -text "max: [format %6g $max_y]" -fill black -tag info -anchor nw
	$w.c create text [expr $width-120] 30 -text "spread: [format %6g [expr $max_y-$min_y]]" -fill black -tag info -anchor nw
	if { $count_y>0 } {
		if { [catch {
			set var [expr $sum_yy*$count_y-$sum_y*$sum_y]
			if { $var < 0 } { set $var 0.0 }
			$w.c create text [expr $width-120] 40 -text "mean: [format %6g [expr $sum_y/$count_y]]" -fill black -tag info -anchor nw
			$w.c create text [expr $width-120] 50 -text "sd: [format %6g [expr sqrt($var)/$count_y]]" -fill black -tag info -anchor nw
			} ] } { puts "$sum_y $sum_yy $count_y" }
		} {
		$w.c create text [expr $width/2] [expr $height/2] -text "No data" -fill red -tag "no_data"
		}
	$w.c create text [expr $width-120] [expr $height-20] -text "min: [format %6g $min_y]" -fill black -tag info -anchor nw
	}
set $w.updating 0
}

proc make_lvdt_display { w } {
global $w.data
set $w.data {}

frame $w -background black
make_graph_widget $w.lvdt1_coarse
make_graph_widget $w.lvdt1_fine
make_graph_widget $w.lvdt2_coarse
make_graph_widget $w.lvdt2_fine

make_graph_widget $w.lvdt1_avg_coarse 300 200 1000 2
make_graph_widget $w.lvdt1_avg_fine 300 200 1000 2
make_graph_widget $w.lvdt2_avg_coarse 300 200 1000 2
make_graph_widget $w.lvdt2_avg_fine 300 200 1000 2

label $w.l_lvdt1_coarse -text "LVDT1 coarse (V)" -foreground yellow -background black
label $w.l_lvdt1_fine -text "LVDT1 fine (V)" -foreground yellow -background black
label $w.l_lvdt2_coarse -text "LVDT2 coarse (V)" -foreground yellow -background black
label $w.l_lvdt2_fine -text "LVDT2 fine (V)" -foreground yellow -background black

grid $w.l_lvdt1_coarse $w.l_lvdt2_coarse $w.l_lvdt1_fine $w.l_lvdt2_fine
grid $w.lvdt1_coarse $w.lvdt2_coarse $w.lvdt1_fine $w.lvdt2_fine -padx 1 -pady 1
grid $w.lvdt1_avg_coarse $w.lvdt2_avg_coarse $w.lvdt1_avg_fine $w.lvdt2_avg_fine -padx 1 -pady 1
}

proc add_lvdt_data { w data } {
global $w.data

foreach {t aux0 aux1 aux2 aux3 x y z s} $data {}
add_graph_point $w.lvdt1_coarse $t $x
add_graph_point $w.lvdt1_fine $t $y
add_graph_point $w.lvdt2_coarse $t $z
add_graph_point $w.lvdt2_fine $t $s

lappend $w.data $t $x $y $z $s
if { [llength [set $w.data]] > 50 } {
	foreach var {st sx sy sz ss} {
		set $var 0.0
		}
	set i 0
	foreach {t x y z s} [set $w.data] {
		foreach var {t x y z s} {
			catch {
				set s$var [expr [set s$var]+[set $var]]
				}
			}
		incr i
		}
	foreach var {t x y z s} {
		set s$var [expr [set s$var]/$i]
		}

	add_graph_point $w.lvdt1_avg_coarse $st $sx
	add_graph_point $w.lvdt1_avg_fine $st $sy
	add_graph_point $w.lvdt2_avg_coarse $st $sz
	add_graph_point $w.lvdt2_avg_fine $st $ss

	set $w.data {}
	}
}

proc make_control_widget { w data } {
frame $w

grid columnconfig $w 1 -weight 1

set i 0
foreach {type desc var extra1 extra2 } $data {
	switch -exact -- $type {
		heading {
			label $w.l$i -text $desc -background black -foreground yellow
			grid $w.l$i - - -sticky news
			}
		entry {
			global $var
			label $w.l$i -text $desc
			entry $w.e$i -textvariable $var
			grid $w.l$i $w.e$i -sticky news
			}
		action_entry {
			global $var
			label $w.l$i -text $desc
			entry $w.e$i -textvariable $var
			button $w.b$i -text $extra2 -command $extra1
			grid $w.l$i $w.e$i $w.b$i -sticky news
			}
		variable_entry {
			global $var
			label $w.l$i -text $desc
			entry $w.e$i -textvariable $var
			button $w.b$i -text "Set" -command "send_cmd \"set_variable $var \[set $var\]\""
			grid $w.l$i $w.e$i $w.b$i -sticky news
			}
		command {
			label $w.l$i -text $desc
			button $w.b$i -text $extra2 -command $extra1
			grid $w.l$i $w.b$i - -sticky news
			}
		default {
			puts stderr "** INTERNAL ERROR: unknown type $type for {$type $desc $var}"
			}
		}
	incr i
	}
}

make_lvdt_display .g1
grid .g1 - -


#grid columnconfig 0 -weight 1

make_phase_widget .phases
grid .phases - -

make_control_widget .control1 {
	heading "Data" "" "" ""
	entry "Timer" "timestamp" "" ""
	entry "Samples acquired" "sample_count" "" ""
	entry "LVDT1 coarse" "lvdt1_coarse" "" ""
	entry "LVDT2 coarse" "lvdt2_coarse" "" ""
	entry "LVDT1 fine" "lvdt1_fine" "" ""
	entry "LVDT2 fine" "lvdt2_fine" "" ""
	entry "Position" "position" "" ""
	entry "Working point deviation" "working_point_mad" "" ""
	entry "Injection value"  "injection" "" ""
	entry "Mark"  "mark" "" ""
	heading "Outputs" "" "" ""
	entry "dac0" "dac0" "" ""
	entry "dac1" "dac1" "" ""
	entry "dac2" "dac2" "" ""
	entry "dac3" "dac3" "" ""
        action_entry "LEDs" "led_value" {send_cmd "set_leds [expr $led_value+256]"} "Set"
	}

make_control_widget .control2 {
	heading "Injection control" "" "" ""
	variable_entry "Injection mask" "injection_mask" "" ""
	variable_entry "Sine frequency" "sine_frequency" "" ""
	variable_entry "Sine phase" "sine_phase" "" ""
	variable_entry "Sine amplitude" "sine_amplitude" "" ""
	variable_entry "Noise amplitude" "noise_amplitude" "" ""
	variable_entry "Noise amplitude" "noise_amplitude" "" ""
	variable_entry "Cycle burst amplitude" "cycle_burst_amplitude" "" ""
	variable_entry "Cycle burst frequency" "cycle_burst_frequency" "" ""
	variable_entry "Cycle burst count" "cycle_burst_max_cycles" "" ""
	}

#       command "" "" {send_cmd "set_dac4x 2 $dac0_value $dac1_value $dac2_value $dac3_value"} "Set DAC values"


make_control_widget .control3 {
	heading "Feedback control" "" "" ""
        variable_entry "control mode" "control_mode" "" ""
	variable_entry "working point" "working_point" "" ""
	variable_entry "kick constant" "kick_constant" "" ""
	variable_entry "damping constant" "damping_constant" "" ""
        variable_entry "damping highpass" "damping_highpass" "" ""
        variable_entry "damping slew rate" "damping_slew_rate" "" ""
	variable_entry "integrator slew rate" "integrator_slew_rate" "" ""
	variable_entry "integrator bang strength" "integrator_bang_strength" "" ""
	variable_entry "antispring" "antispring" "" ""
	heading "System controls" "" "" ""
	command "Reinialize DAQ board" "" {send_cmd reinitialize} "Reinit"
	command "Stop DAQ board" "" {send_cmd stop_board} "Stop"
	}

grid .control1 .control2 .control3 -sticky new -padx 2

update

fconfigure $CONTROL -blocking false
fileevent $CONTROL readable "read_data $CONTROL"

send_cmd "return_vars"

