#!/usr/bin/env wish

foreach {var value} {
	line_count 0
	dac0_value 0.0
	dac1_value 0.0
	dac2_value 0.0
	dac3_value 0.0
	led_value 0
	} { 
	global $var
	set $var $value
	}

global CONTROL
set CONTROL [socket "localhost" 35546]

proc wrap_up {} {
global CONTROL
close $CONTROL
exit 0
}

wm protocol . WM_DELETE_WINDOW wrap_up
wm title . "General purpose DAQ interface"

global LOG
set LOG [open "capture.log" "w"]

proc read_data { c } {
global LOG

fconfigure $c -blocking true
gets $c line
fconfigure $c -blocking false

if { $line == "" } { return }

switch -exact -- [lindex $line 0] {
	PHASE   {
		set_phase_row .phases [lindex $line 1] [lrange $line 2 end]
		}

	JEEP1 {
		puts $LOG $line
		set mean [lrange $line 4 11]
		set max [lrange $line 13 20]
		set min [lrange $line 22 30]
		set phases [lrange $line 31 47]
		add_adc_data .g1 [lindex $line 2] $min $mean $max
		add_phase_data .phases [lindex $line 2] $phases
		global daemon_timer samples_acquired
		foreach {daemon_timer samples_acquired} [lrange $line 1 2] {}
		for { set i 0 } { $i < 8 } { incr i } {
			global adc$i
			set adc$i [lindex $mean $i]
			}
		}

	default {
		puts $LOG $line
		#.line delete 0 end
		#.line insert 0 $line
		}
	}

}

proc send_cmd { s } {
global CONTROL
fconfigure $CONTROL -blocking true
puts $CONTROL $s
flush $CONTROL
fconfigure $CONTROL -blocking false
}

proc make_phase_widget1 { w } {

frame $w

for { set i 0 } { $i < 8 } { incr i } {
	label $w.l$i -text "ch$i"
	grid $w.l$i -row [expr $i+1] -column 0 -sticky news
	}

for { set j 0 } { $j < 8 } { incr j } {
	label $w.lp$j -text "Phase $j"
	grid $w.lp$j -column [expr $j+1] -row 0 -sticky news
	}

for { set i 0 } { $i < 8 } { incr i } {
	for { set j 0 } { $j < 8 } { incr j } {
		entry $w.phase_${i}_$j
		grid $w.phase_${i}_$j -row [expr $i+1] -column [expr $j+1] -sticky news
		}
	}
}

proc set_phase_row { w row data } {
for { set i 0 } { $i < 8 } { incr i } {
	$w.phase_${row}_$i delete 0 end
	$w.phase_${row}_$i insert 0 [lindex $data [expr $i]]
	}
}

proc make_graph_widget { w {width 300} { height 180} {max_points 250} {incr_points 20} } {
frame $w
canvas $w.c -width $width -height $height -background white
grid $w.c

global $w.data $w.settings $w.updating
set $w.data {}
set $w.updating 0

foreach var {max_points incr_points} {
	set $w.settings($var) [set $var]
	}

$w.c create text [expr $width/2] [expr $height/2] -text "No data" -fill red -tag "no_data"
}

proc add_graph_point { w x y } {
global $w.data $w.settings $w.updating

lappend $w.data $x $y

foreach var { max_points incr_points } {
	set $var [set $w.settings($var)]
	}

if { [llength [set $w.data]] > $max_points+$incr_points } { 
	set $w.data [lrange [set $w.data] $incr_points end]

	if { ! [set $w.updating] } {
		set $w.updating 1
		after idle "update_graph $w"
		}
	}

if { [llength [set $w.data]] <= $max_points && ([llength [set $w.data]] % $incr_points==0) } {
	if { ! [set $w.updating] } {
		set $w.updating 1
		after idle "update_graph $w"
		}
	}

}


proc update_graph { w } {
global $w.data $w.settings $w.updating $w.updating

if { [llength [set $w.data]] >= 4 } {
	set x0 [lindex [set $w.data] 0]
	set y0 [lindex [set $w.data] 1]

	set max_x -1e10
	set min_x 1e10
	set max_y -1e10
	set min_y 1e10
	set sum_y 0.0
	set sum_yy 0.0
	set count_y 0

	foreach {x y } [set $w.data] {
		if { $y== "Inf" || $y == "-Inf" || "$y" == "nan" || "$y" == "-nan" || $y == "inf" || $y== "-inf"} { continue }
		if { $x > $max_x } { set max_x $x }
		if { $x < $min_x } { set min_x $x }
		if { $y > $max_y } { set max_y $y }
		if { $y < $min_y } { set min_y $y }

		if { [catch {
		set sum_y [expr $sum_y+$y]
			} ]} { puts "x $sum_y $sum_yy $y" }
		incr count_y
		}

	catch {
		set mean [expr $sum_y/$count_y]

		set sum_yy 0.0
		set count_y 0
		foreach {x y } [set $w.data] {
			set sum_yy [expr $sum_yy+($y-$mean)*($y-$mean)]
			incr count_y
			}
		}

	if { $max_y <= $min_y } { 
		set max_y [expr $min_y+1]
		set min_y [expr $min_y-1]
		}

	set width [$w.c cget -width]
	set height [$w.c cget -height]
	set inv_x 0.0
	set inv_y 0.0
	catch {
		set inv_x [expr $width/($max_x-$min_x+1e-56)]
		set inv_y [expr $height/($max_y-$min_y+1e-56)]
		}


	if { $max_x < $min_x || $max_y < $min_y } { 
		set $w.updating 0
		return 
		}

	$w.c delete no_data
	$w.c delete curve

	foreach {x y } [lrange [set $w.data] 2 end] {
		if { $y== "Inf" || $y == "-Inf" } { continue }
		catch {
			$w.c create line [expr ($x0-$min_x)*$inv_x] [expr ($max_y-$y0)*$inv_y] [expr ($x-$min_x)*$inv_x] [expr ($max_y-$y)*$inv_y] -fill blue -tag curve
			}
		set x0 $x
		set y0 $y
		}

	$w.c delete info
	$w.c create text [expr $width-120] 15 -text "max: [format %.7f $max_y]" -fill black -tag info -anchor nw
	$w.c create text [expr $width-120] 25 -text "spread: [format %g [expr $max_y-$min_y]]" -fill black -tag info -anchor nw
	if { $count_y>0 } {
		if { [catch {
			$w.c create text [expr $width-120] 35 -text "mean: [format %.7f [expr $sum_y/$count_y]]" -fill black -tag info -anchor nw
			$w.c create text [expr $width-120] 45 -text "sd: [format %g [expr sqrt($sum_yy)/$count_y]]" -fill black -tag info -anchor nw
			} ] } { puts "y $sum_y $sum_yy $count_y" }
		} {
		$w.c create text [expr $width/2] [expr $height/2] -text "No data" -fill red -tag "no_data"
		}
	$w.c create text [expr $width-120] [expr $height-20] -text "min: [format %.7f $min_y]" -fill black -tag info -anchor nw
	}
set $w.updating 0
}

proc make_adc_display { w } {

frame $w -background black
for { set i 0 } { $i < 8 } { incr i } {
	label $w.l_adc${i} -text "ADC $i" -foreground yellow -background black
	make_graph_widget $w.adc${i}
	make_graph_widget $w.adc_spread${i}
	#make_graph_widget $w.adc${i}_avg 300 200 1000 2
	grid $w.l_adc${i} -column [expr $i % 4] -row [expr 0+3*($i/4)] -padx 1
	grid $w.adc${i} -column [expr $i% 4] -row [expr 1+3*($i/4)] -padx 1 -pady 2
	grid $w.adc_spread${i} -column [expr $i% 4] -row [expr 2+3*($i/4)] -padx 1 -pady 2
	#grid $w.adc${i}_avg -column [expr $i% 4] -row [expr 2+3*($i/4)] -padx 1
	}
}

proc add_adc_data { w t min mean max} {

for { set i 0 } { $i < 8 } { incr i } {
	add_graph_point $w.adc${i} $t [lindex $mean $i]
	add_graph_point $w.adc_spread${i} $t [expr [lindex $max $i]-[lindex $min $i]]
	}
}

proc make_phase_display { w } {

frame $w -background black
for { set i 0 } { $i < 16 } { incr i } {
	label $w.l_phase${i} -text "PHASE $i" -foreground yellow -background black
	make_graph_widget $w.phase${i} 200 80
	grid $w.l_phase${i} -column [expr $i % 2] -row [expr 0+2*($i/2)] -padx 1
	grid $w.phase${i} -column [expr $i% 2] -row [expr 1+2*($i/2)] -padx 1 -pady 2
	#grid $w.adc${i}_avg -column [expr $i% 4] -row [expr 2+3*($i/4)] -padx 1
	}
}

proc add_phase_data { w t phases} {

for { set i 0 } { $i < 16 } { incr i } {
	add_graph_point $w.phase${i} $t [lindex $phases $i]
	}
}

proc make_control_widget { w data } {
frame $w

grid columnconfig $w 1 -weight 1

set i 0
foreach {type desc var extra1 extra2 } $data {
	switch -exact -- $type {
		heading {
			label $w.l$i -text $desc -background black -foreground yellow
			grid $w.l$i - - -sticky news
			}
		entry {
			global $var
			label $w.l$i -text $desc
			entry $w.e$i -textvariable $var
			grid $w.l$i $w.e$i -sticky news
			}
		action_entry {
			global $var
			label $w.l$i -text $desc
			entry $w.e$i -textvariable $var
			button $w.b$i -text $extra2 -command $extra1
			grid $w.l$i $w.e$i $w.b$i -sticky news
			}
		command {
			label $w.l$i -text $desc
			button $w.b$i -text $extra2 -command $extra1
			grid $w.l$i $w.b$i - -sticky news
			}
		default {
			puts stderr "** INTERNAL ERROR: unknown type $type for {$type $desc $var}"
			}
		}
	incr i
	}
}

make_adc_display .g1
make_phase_display .phases
grid .g1 .phases


#grid columnconfig 0 -weight 1

make_control_widget .control1 {
	heading "Data" "" "" ""
	entry "Timer" "daemon_timer" "" ""
	entry "Samples acquired" "samples_acquired" "" ""
	entry "adc0" "adc0" "" ""
	entry "adc1" "adc1" "" ""
	entry "adc2" "adc2" "" ""
	entry "adc3" "adc3" "" ""
	entry "adc4" "adc4" "" ""
	entry "adc5" "adc5" "" ""
	entry "adc6" "adc6" "" ""
	entry "adc7" "adc7" "" ""
	}

make_control_widget .control2	{
	heading "Outputs" "" "" ""
	entry "dac0" "dac0_value" "" ""
	entry "dac1" "dac1_value" "" ""
	entry "dac2" "dac2_value" "" ""
	entry "dac3" "dac3_value" "" ""
        command "" "" {send_cmd "set_dac4x 2 $dac0_value $dac1_value $dac2_value $dac3_value"} "Set DAC values"
        action_entry "LEDs" "led_value" {send_cmd "set_leds [expr $led_value+256]"} "Set"
	heading "System controls" "" "" ""
	command "Reinialize DAQ board" "" {send_cmd reinitialize} "Reinit"
	command "Stop DAQ board" "" {send_cmd stop_board} "Stop"
	}

grid .control1 .control2 -sticky new -padx 2

update

fconfigure $CONTROL -blocking false
fileevent $CONTROL readable "read_data $CONTROL"

send_cmd "set_leds 256"
