#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h> /* the L2 protocols */
#define _GNU_SOURCE
#define __USE_GNU
#include <sched.h>
#include "daemon.h"
#include "receive_packet.h"
#include "control_interface.h"
#include "tiltmeter.h"
#include "jeep1.h"
#include "seism1.h"
#include "cmdline.h"

extern FILE * data_log;
extern struct gengetopt_args_info args_info;
extern int board_initialized;

typedef struct {
	unsigned char dst_mac[6];
	unsigned char src_mac[6];
	unsigned char type[2];
	unsigned char packet_count[6];
	} ETHERNET_DATA_HEADER;

typedef struct {
	unsigned char type[2];
	unsigned char sample_count[6];
	unsigned char channel_data[12];
	} SUBPACKET_HEADER;
	
	
typedef struct {
	unsigned char type[2];
	unsigned char switches[2];
	unsigned tx_counter;
	u64 general_config;
	u64 port_config;
	unsigned char source_mac[6];
	unsigned char filler1[2];
	unsigned char destination_mac[6];
	unsigned char filler2[2];
	unsigned sclk_counter;	
	} __attribute__((__packed__)) STATUS_SUBPACKET;


u64 count6(unsigned char *buf)
{
u64 a;
int i;
a=buf[5];

for(i=4;i>=0;i--)a=(a<<8)+buf[i];
return(a);
}

double channel_sample(unsigned char *buf)
{
int word;
double a;
return 0;
word=(buf[2] & 0x80) ? -1 : 0;
memcpy(&word, buf, 3);
if((word & 0xffffff) ==  0x7fffff)return(INFINITY);
if((word & 0xffffff) ==  0x800000)return(-INFINITY);
a= (word*2.5)/((1<<23)-1.0);
return(a);
}

long long prev_subpacket_count[16]={-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

void parse_subpacket(ETHERNET_DATA_HEADER *ethernet_header, unsigned char *buffer, int len)
{
SUBPACKET_HEADER *subpacket_header=buffer;
STATUS_SUBPACKET *status=buffer;
unsigned short *sw;
double a;
double adc[8];
int i;
int channel=subpacket_header->type[0] & 0xf;
int phases=subpacket_header->type[1] & 0xf;

//fprintf(stderr, "type=%02x%02x len=%d %d %1x\n", subpacket_header->type[0], subpacket_header->type[1], len, channel, phases);
switch(subpacket_header->type[1] & 0xf0) {
	case 0x70: {
		if(!board_initialized)break;
		
		a=count6(buffer+2);
		if( (a!=prev_subpacket_count[channel]+1) && (prev_subpacket_count[channel]>0)) {
			fprintf(stdout, "Missed subpacket: %.0f vs %.0f\n", prev_subpacket_count[channel], a);
			}
		prev_subpacket_count[channel]=a;
		

		if(data_log!=NULL) {
			fprintf(data_log, "SENSE %02x:%02x:%02x:%02x:%02x:%02x %d %d %.0f", 
			ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
			subpacket_header->type[0],
			subpacket_header->type[1],
			a);


			sw=buffer+8;
			
			for(i=0;i<6;i++)fprintf(data_log, " %d", sw[i]);
			fprintf(data_log, "\n");
			}

// 		broadcast("SENSE %02x:%02x:%02x:%02x:%02x:%02x %d %d %.0f", 
// 			ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
// 			subpacket_header->type[0],
// 			subpacket_header->type[1],
// 			a);
// 		sw=buffer+8;
// 		
// 		for(i=0;i<6;i++)broadcast(" %d", sw[i]);
// 		broadcast("\n");

		break;
		}
	case 0x40: {
		if(!board_initialized)break;
/*		fprintf(stderr, "%d %f [%.6f %.6f %.6f %.6f]\n", subpacket_header->type[0],
			count6(buffer+2),
			channel_sample(buffer+8),
			channel_sample(buffer+11),
			channel_sample(buffer+14),
			channel_sample(buffer+17));*/

		a=count6(buffer+2);
		if( (a!=prev_subpacket_count[channel]+1) && (prev_subpacket_count[channel]>0)) {
			fprintf(stderr, "Missed subpacket: %.0f vs %.0f\n", prev_subpacket_count[channel], a);
			}
		prev_subpacket_count[channel]=a;


		if(data_log!=NULL)fprintf(data_log, "ADC %02x:%02x:%02x:%02x:%02x:%02x %d %d %.0f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f\n", 
			ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
			subpacket_header->type[0],
			subpacket_header->type[1],
			a,
			channel_sample(buffer+8),
			channel_sample(buffer+11),
			channel_sample(buffer+14),
			channel_sample(buffer+17),
			channel_sample(buffer+20),
			channel_sample(buffer+23),
			channel_sample(buffer+26),
			channel_sample(buffer+29));

		for(i=0;i<8;i++)adc[i]=channel_sample(buffer+8+3*i);
			
		if(args_info.tiltmeter_mode_arg)tiltmeter_process_adc_packet(channel, phases, a, adc, 8);
		if(args_info.jeep1_mode_arg)jeep1_process_adc_packet(channel, phases, a, adc, 8);
		if(args_info.seism1_mode_arg)seism1_process_adc_packet(channel, phases, a, adc, 8);

				    
/*		broadcast("ADC %02x:%02x:%02x:%02x:%02x:%02x %d %d %.0f %.8f %.8f %.8f %.8f\n", 
			ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
			subpacket_header->type[0],
			subpacket_header->type[1],
			a,
			channel_sample(buffer+8),
			channel_sample(buffer+11),
			channel_sample(buffer+14),
			channel_sample(buffer+17));*/
		break;
		}
	case 0x80: {
		if(data_log!=NULL) {
			fprintf(data_log, "STATUS %02x:%02x:%02x:%02x:%02x:%02x %d %d %x %u %u %016lx %016lx\n", 
				ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
				
				subpacket_header->type[0],
				subpacket_header->type[1],
				
				status->switches[0],
				status->tx_counter,
				status->sclk_counter,
				status->general_config,
				status->port_config);
			for(i=0;i<len;i++)fprintf(data_log, " %02x", buffer[i]);
			fprintf(data_log, "\n");
			}
		broadcast("STATUS %02x:%02x:%02x:%02x:%02x:%02x %d %d %02x %02x %u %u %016lx %016lx\n", 
			ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
			
			subpacket_header->type[0],
			subpacket_header->type[1],
			
			status->switches[0],
			status->switches[1],
			status->tx_counter,
			status->sclk_counter,
			status->general_config,
			status->port_config);
/*		for(i=0;i<len;i++)broadcast(" %02x", buffer[i]);
		broadcast("\n");*/
		break;
		}
	default: {
		if(data_log!=NULL) {
			fprintf(data_log, "UNKNOWN PACKET %02x:%02x:%02x:%02x:%02x:%02x %d %d %d", 
				ethernet_header->src_mac[0], ethernet_header->src_mac[1], ethernet_header->src_mac[2], ethernet_header->src_mac[3], ethernet_header->src_mac[4], ethernet_header->src_mac[5],
				subpacket_header->type[0],
				subpacket_header->type[1], len);
			for(i=0;i<len;i++)fprintf(data_log, " %02x", buffer[i]);
			fprintf(data_log, "\n");
			}
		}
	
	}

}

double prev_ethernet_packet_count=-1;

void parse_packet(unsigned char *buffer, int len)
{
ETHERNET_DATA_HEADER *header=buffer;
unsigned short *packet_size;
int i;
double a;

a=count6(header->packet_count);
if( (a!=prev_ethernet_packet_count+1) && (prev_ethernet_packet_count>0)) {
	fprintf(stderr, "Missed packet: %.0f vs %.0f\n", prev_ethernet_packet_count, a);
	}
prev_ethernet_packet_count=a;
return;
//fprintf(stderr, "len=%d\n", len);
// if(len!=351) {
// 	for(i=0;i<len;i++)fprintf(stderr, " %02x", buffer[i]);
// 	fprintf(stderr, "\n");
// 	}

i=sizeof(ETHERNET_DATA_HEADER);
while(i+1<len) {
	packet_size=&(buffer[i]);
/*	if(*packet_size!=20)fprintf(stderr, "%02x:%02x:%02x:%02x:%02x:%02x [%d %d %d]\n",
		header->src_mac[0], header->src_mac[1], header->src_mac[2], header->src_mac[3], header->src_mac[4], header->src_mac[5], i, len, *packet_size);*/
/*	if(*packet_size==0) {
		fprintf(stderr, "%02x %02x %02x %02x %02x %02x %02x %02x\n",
			buffer[i], buffer[i+1], buffer[i+2], buffer[i+3], buffer[i+4], buffer[i+5], buffer[i+6], buffer[i+7]);
		}*/
	if(i+2+*packet_size<len)parse_subpacket(header, buffer+i+2, *packet_size);
		else break;
	i+=2+*packet_size;
	}
	
//if(header->packet_count[0]==0)fflush(data_log);
}
