#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h> /* the L2 protocols */
#include <errno.h>
#include <math.h>
#include "registers.h"
#include "send_cmd.h"
#include "daemon.h"
#include "cmdline.h"

extern struct gengetopt_args_info args_info;

ETHERNET_HEADER test_header={
	{0x02, 0x89, 0x67, 0x45, 0x23, 0x01},
	{0x00, 0x24, 0xe8, 0x93, 0x99, 0x6c},
	{0xda, 0x8a}
//	{0x8a, 0xda}
	};

extern int raw_sock;

int board_initialized=0;


void fill_crc32(unsigned char *packet, int len)
{
char CrcNext[32];
char Data[4];
char Crc[32];
unsigned int crc_hex, data_hex;
int i, j;
int Enable=1;

for(i=0;i<32;i++)Crc[i]=1;
	
for(i=0;i<(2*(len-4));i++) {
	
	if((i<(2*(len-4))) && (i>=0)) {
		Data[0]=1 & (packet[i>>1]>>(4*(i &1)+3));
		Data[1]=1 & (packet[i>>1]>>(4*(i &1)+2));
		Data[2]=1 & (packet[i>>1]>>(4*(i &1)+1));
		Data[3]=1 & (packet[i>>1]>>(4*(i &1)+0));
		Enable=1;
		} else {
		Data[0]=0;
		Data[0]=0;
		Data[0]=0;
		Data[0]=0;
		Enable=0;
		}
	
	CrcNext[0] = Enable & (Data[0] ^ Crc[28]); 
	CrcNext[1] = Enable & (Data[1] ^ Data[0] ^ Crc[28] ^ Crc[29]); 
	CrcNext[2] = Enable & (Data[2] ^ Data[1] ^ Data[0] ^ Crc[28] ^ Crc[29] ^ Crc[30]); 
	CrcNext[3] = Enable & (Data[3] ^ Data[2] ^ Data[1] ^ Crc[29] ^ Crc[30] ^ Crc[31]); 
	CrcNext[4] = (Enable & (Data[3] ^ Data[2] ^ Data[0] ^ Crc[28] ^ Crc[30] ^ Crc[31])) ^ Crc[0]; 
	CrcNext[5] = (Enable & (Data[3] ^ Data[1] ^ Data[0] ^ Crc[28] ^ Crc[29] ^ Crc[31])) ^ Crc[1]; 
	CrcNext[6] = (Enable & (Data[2] ^ Data[1] ^ Crc[29] ^ Crc[30])) ^ Crc[ 2]; 
	CrcNext[7] = (Enable & (Data[3] ^ Data[2] ^ Data[0] ^ Crc[28] ^ Crc[30] ^ Crc[31])) ^ Crc[3]; 
	CrcNext[8] = (Enable & (Data[3] ^ Data[1] ^ Data[0] ^ Crc[28] ^ Crc[29] ^ Crc[31])) ^ Crc[4]; 
	CrcNext[9] = (Enable & (Data[2] ^ Data[1] ^ Crc[29] ^ Crc[30])) ^ Crc[5]; 
	CrcNext[10] = (Enable & (Data[3] ^ Data[2] ^ Data[0] ^ Crc[28] ^ Crc[30] ^ Crc[31])) ^ Crc[6]; 
	CrcNext[11] = (Enable & (Data[3] ^ Data[1] ^ Data[0] ^ Crc[28] ^ Crc[29] ^ Crc[31])) ^ Crc[7]; 
	CrcNext[12] = (Enable & (Data[2] ^ Data[1] ^ Data[0] ^ Crc[28] ^ Crc[29] ^ Crc[30])) ^ Crc[8]; 
	CrcNext[13] = (Enable & (Data[3] ^ Data[2] ^ Data[1] ^ Crc[29] ^ Crc[30] ^ Crc[31])) ^ Crc[9]; 
	CrcNext[14] = (Enable & (Data[3] ^ Data[2] ^ Crc[30] ^ Crc[31])) ^ Crc[10]; 
	CrcNext[15] = (Enable & (Data[3] ^ Crc[31])) ^ Crc[11]; 
	CrcNext[16] = (Enable & (Data[0] ^ Crc[28])) ^ Crc[12]; 
	CrcNext[17] = (Enable & (Data[1] ^ Crc[29])) ^ Crc[13]; 
	CrcNext[18] = (Enable & (Data[2] ^ Crc[30])) ^ Crc[14]; 
	CrcNext[19] = (Enable & (Data[3] ^ Crc[31])) ^ Crc[15]; 
	CrcNext[20] = Crc[16]; 
	CrcNext[21] = Crc[17]; 
	CrcNext[22] = (Enable & (Data[0] ^ Crc[28])) ^ Crc[18]; 
	CrcNext[23] = (Enable & (Data[1] ^ Data[0] ^ Crc[29] ^ Crc[28])) ^ Crc[19]; 
	CrcNext[24] = (Enable & (Data[2] ^ Data[1] ^ Crc[30] ^ Crc[29])) ^ Crc[20]; 
	CrcNext[25] = (Enable & (Data[3] ^ Data[2] ^ Crc[31] ^ Crc[30])) ^ Crc[21]; 
	CrcNext[26] = (Enable & (Data[3] ^ Data[0] ^ Crc[31] ^ Crc[28])) ^ Crc[22]; 
	CrcNext[27] = (Enable & (Data[1] ^ Crc[29])) ^ Crc[23]; 
	CrcNext[28] = (Enable & (Data[2] ^ Crc[30])) ^ Crc[24]; 
	CrcNext[29] = (Enable & (Data[3] ^ Crc[31])) ^ Crc[25]; 
	CrcNext[30] = Crc[26]; 
	CrcNext[31] = Crc[27]; 
	
	data_hex=(Data[3]<<3)+(Data[2]<<2)+(Data[1]<<1)+Data[0];
	
	crc_hex=0;
	for(j=0;j<32;j++) {
		Crc[j]=CrcNext[j];
		crc_hex=(crc_hex<<1) | !CrcNext[j];
		}
	//printf("%d %01x %08x (inv: %08x) %d %d %d %d\n", i, data_hex, crc_hex, ~crc_hex, !CrcNext[31], !CrcNext[30],!CrcNext[29],!CrcNext[28]);
	}
	
packet[len-4]=crc_hex & 0xff;
packet[len-3]=(crc_hex >>8) & 0xff;
packet[len-2]=(crc_hex >>16)& 0xff;
packet[len-1]=(crc_hex >>24) & 0xff;

fflush(stdout);
}

#define MAKE_CMD(cmd, count, address)   ( ((u64)(cmd & 0xff)<<56) | ((u64)(count & 0xffff)<<40) | (address & 0xffffffff))

#define PHASE_CLOCK_DIVIDER args_info.phase_clock_divider_arg
#define PHASE_CLOCK_PHASE args_info.phase_clock_phase_arg

#define MAX_BUFFER_LEN 128
unsigned char out_buffer[MAX_BUFFER_LEN*8+sizeof(ETHERNET_HEADER)+8];
u64 *out_cmd_buffer=(u64 *)(out_buffer+sizeof(ETHERNET_HEADER));
int out_cmd_free=0;
struct sockaddr_ll sock_header;

void parse_mac(char *str_mac, unsigned char *mac)
{
while(*str_mac && (*str_mac==' ' || *str_mac=='\t'))str_mac++;
sscanf(str_mac, "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx", &(mac[0]), &(mac[1]), &(mac[2]), &(mac[3]), &(mac[4]), &(mac[5]));
}

void init_out_buffer(void)
{
ETHERNET_HEADER *header=out_buffer;

parse_mac(args_info.board_mac_arg, test_header.dst_mac);
parse_mac(args_info.daemon_mac_arg, test_header.src_mac);

// fprintf(stderr, "%02x:%02x:%02x:%02x:%02x:%02x", (test_header.src_mac[0]), (test_header.src_mac[1]), (test_header.src_mac[2]), (test_header.src_mac[3]), (test_header.src_mac[4]), (test_header.src_mac[5]));

memset(&sock_header, 0, sizeof(sock_header));

sock_header.sll_family=AF_PACKET;
sock_header.sll_protocol=0x8ada;
sock_header.sll_ifindex=2; /* use interface 1 */
sock_header.sll_pkttype=PACKET_HOST;
sock_header.sll_halen=6;
memcpy(sock_header.sll_addr, test_header.dst_mac, 6);
memcpy(header, &test_header, sizeof(*header));

out_cmd_free=0;
}

void flush_out_buffer(void)
{
int i, len;

if(out_cmd_free<1)return;

do {
	out_cmd_buffer[out_cmd_free++]=MAKE_CMD(1, 1, REG_NOP);
	out_cmd_buffer[out_cmd_free++]=0;
	} while(out_cmd_free<8);

len=sizeof(ETHERNET_HEADER)+8*out_cmd_free;
//fprintf(stderr, "Buffer length is %d\n", len);

fill_crc32(out_buffer, len);

while((i=sendto(raw_sock, out_buffer, len, 0, &sock_header, sizeof(sock_header)))<0) {
	if(i!=ENOBUFS) {
		perror("Error sending message");
		fprintf(stderr, "errno=%d\n", i);
		exit(-1);
		}
	}

out_cmd_free=0;
}

#define SET_REGISTER(reg, value)    { \
	if(out_cmd_free+2>=MAX_BUFFER_LEN)flush_out_buffer(); \
	out_cmd_buffer[out_cmd_free++]=MAKE_CMD(CMD_WRITE_RX_REG, 1, reg); \
	out_cmd_buffer[out_cmd_free++]=value; \
	}
	
#define NOP {\
	SET_REGISTER(REG_NOP, 0); \
	}
	
#define CONTROL_CMD_START(count) {\
	if(out_cmd_free+1+count>=MAX_BUFFER_LEN)flush_out_buffer(); \
	out_cmd_buffer[out_cmd_free++]=MAKE_CMD(CMD_WRITE_COMMAND, count, 0); \
	}
	
#define CONTROL_CMD(cmd) {\
	out_cmd_buffer[out_cmd_free++]=cmd; \
	}

#define NOP_CMD {\
	CONTROL_CMD(0); \
	}

#define DAC8734_CMD(port, reg, value) {\
	CONTROL_CMD( (((u64)(port & 0x3))<<60) | (((u64)(0x4))<<56) | ((reg & 0xff)<<16) | (value & 0xffff)); \
	}

#define DAC8734_OUTPUT_CMD(port, dac, value, enable) {\
	CONTROL_CMD( (((u64)(port & 0x3))<<60) | (((u64)(dac & 0x3))<<56) | ((enable & 0x1)<<16) | (value & 0xffff)); \
	}

#define DAC1220_CMD(port, reg, count, value) {\
	CONTROL_CMD( (((u64)(port & 0x3))<<60) | (((u64)(0x4))<<56) | (((u64)0x00 | (((u64)count & 0x03)<<5) | ((u64)reg & 0x0f))<<24) | ((u64)value & 0xffffff)); \
	}

void init_board1(void)
{
u64 daemon_mac;
int i;

daemon_mac=0;
for(i=0;i<6;i++)daemon_mac|=(u64)(test_header.src_mac[i] & 0xff)<<(i*8);
fprintf(stderr, "Board destination address %012lx\n", daemon_mac);

SET_REGISTER(REG_DESTINATION_MAC, daemon_mac);
	
SET_REGISTER(REG_PORT_CONFIG, 
	     ((args_info.portA_mode_arg & 0xff)<<REG_PORT_CONFIG_SHIFT_PORTA) | 
	     ((args_info.portB_mode_arg & 0xff)<<REG_PORT_CONFIG_SHIFT_PORTB) |
	     ((args_info.portC_mode_arg & 0xff)<<REG_PORT_CONFIG_SHIFT_PORTC) |
	     ((u64)(args_info.portA_enable_sampling_clock_arg & 0x1)<< REG_PORT_CONFIG_SHIFT_PORTA_SAMPLING_CLK_ENABLE) |
	     ((u64)(args_info.portB_enable_sampling_clock_arg & 0x1)<< REG_PORT_CONFIG_SHIFT_PORTB_SAMPLING_CLK_ENABLE) |
	     ((u64)(args_info.portC_enable_sampling_clock_arg & 0x1)<< REG_PORT_CONFIG_SHIFT_PORTC_SAMPLING_CLK_ENABLE) |
	     ((u64)(args_info.portA_nbits_arg & 0xff)<<REG_PORT_CONFIG_SHIFT_PORTA_NBITS) |
	     ((u64)(args_info.portB_nbits_arg & 0xff)<<REG_PORT_CONFIG_SHIFT_PORTB_NBITS) |
	     ((u64)(args_info.portC_nbits_arg & 0xff)<<REG_PORT_CONFIG_SHIFT_PORTC_NBITS));

/* set clock config */
SET_REGISTER(REG_CLOCK_CONFIG,
	        ((long)(args_info.term1_pin0_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN0
	      | ((long)(args_info.term1_pin1_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN1
	      | ((long)(args_info.term1_pin2_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN2
	      | ((long)(args_info.term1_pin3_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN3
	      | ((long)(args_info.term1_pin4_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN4
	      | ((long)(args_info.clkout_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_CLKOUT
	      | ((long)(args_info.clkio_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_CLKIO
	      | (args_info.sampling_clock_arg & 0xf));
	     
if(args_info.sync1_divider_arg>=0)SET_REGISTER(REG_SLOW_CLK1_CONFIG, 
	     REG_SLOW_CLK_CONFIG_STOP | 
	     ((args_info.sync1_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync1_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));

if(args_info.sync2_divider_arg>=0)SET_REGISTER(REG_SLOW_CLK2_CONFIG, 
	     REG_SLOW_CLK_CONFIG_STOP | 
	     ((args_info.sync2_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync2_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));
	     
if(args_info.sync3_divider_arg>=0)SET_REGISTER(REG_SLOW_CLK3_CONFIG, 
	     REG_SLOW_CLK_CONFIG_STOP | 
	     ((args_info.sync3_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync3_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));
	     
if(args_info.sync4_divider_arg>=0)SET_REGISTER(REG_SLOW_CLK4_CONFIG, 
	     REG_SLOW_CLK_CONFIG_STOP | 
	     ((args_info.sync4_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync4_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));

	     
SET_REGISTER(REG_SYNC1_CONFIG, args_info.sync1_config_arg & 0x1f);
SET_REGISTER(REG_SYNC2_CONFIG, args_info.sync2_config_arg & 0x1f);
SET_REGISTER(REG_SYNC3_CONFIG, args_info.sync3_config_arg & 0x1f);
SET_REGISTER(REG_SYNC4_CONFIG, args_info.sync4_config_arg & 0x1f);

SET_REGISTER(REG_A_XIN_DIV, args_info.portA_xin_div_arg);
SET_REGISTER(REG_B_XIN_DIV, args_info.portB_xin_div_arg);
SET_REGISTER(REG_C_XIN_DIV, args_info.portC_xin_div_arg);

flush_out_buffer();
}

void start_board1(void)
{
flush_out_buffer();

SET_REGISTER(REG_CLOCK_CONFIG,
	        ((long)(args_info.term1_pin0_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN0
	      | ((long)(args_info.term1_pin1_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN1
	      | ((long)(args_info.term1_pin2_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN2
	      | ((long)(args_info.term1_pin3_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN3
	      | ((long)(args_info.term1_pin4_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_PIN4
	      | ((long)(args_info.clkout_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_CLKOUT
	      | ((long)(args_info.clkio_arg & 0xf))<<REG_CLOCK_CONFIG_TERM1_SHIFT_CLKIO
	      | (args_info.sampling_clock_arg & 0xf));

if(args_info.sync1_divider_arg>=0)
	SET_REGISTER(REG_SLOW_CLK1_CONFIG, 
	     ((args_info.sync1_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync1_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));

if(args_info.sync2_divider_arg>=0)
	SET_REGISTER(REG_SLOW_CLK2_CONFIG, 
	     ((args_info.sync2_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync2_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));

if(args_info.sync3_divider_arg>=0)
	SET_REGISTER(REG_SLOW_CLK3_CONFIG, 
	     ((args_info.sync3_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync3_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));

if(args_info.sync4_divider_arg>=0)
	SET_REGISTER(REG_SLOW_CLK4_CONFIG, 
	     ((args_info.sync4_divider_arg & 0xffff)<<REG_SLOW_CLK_CONFIG_SHIFT_DIV) |
	     ((args_info.sync4_phase_arg & 0xffff) << REG_SLOW_CLK_CONFIG_SHIFT_PHASE));

	     
SET_REGISTER(REG_RESET_CONFIG, args_info.reset_config_arg);

flush_out_buffer();
}

void init_dac1220(int port)
{
send_dac_cmd(port, 5, 0x80);
}

void calibrate_dac1220(int port)
{
send_dac_cmd(port, 5, 0x81);
}

void sleep_dac1220(int port)
{
send_dac_cmd(port, 5, 0x82);
}

void start_board2(void)
{
SET_REGISTER(REG_RESET_CONFIG, 0);

if(args_info.portA_mode_arg==4) init_dac1220(1);
if(args_info.portB_mode_arg==4) init_dac1220(2);
if(args_info.portC_mode_arg==4) init_dac1220(3);

sleep(1);

if(args_info.portA_mode_arg==4) calibrate_dac1220(1);
if(args_info.portB_mode_arg==4) calibrate_dac1220(2);
if(args_info.portC_mode_arg==4) calibrate_dac1220(3);

sleep(1);

flush_out_buffer();
board_initialized=1;
}

void stop_board(void)
{
SET_REGISTER(REG_PORT_CONFIG, 0);
SET_REGISTER(REG_CLOCK_CONFIG, REG_CLOCK_CONFIG_SAMPLING_CLK_25MHZ);
flush_out_buffer();
}

static int volts_to_DAC8734_code(double value)
{
int x;
x=round(value*3276.80);
if(x>=0x7fff)
	x=0x7fff;
	else
if(x< -32768)
	x=-32768;
return(x & 0xffff);
}

static int volts_to_DAC1220_code(double value)
{
int x;
x=round((value-2.5)*209715.2);
if(x>=0x7ffff)
	x=0x7ffff;
	else
if(x< -524288)
	x=-524288;
return((x & 0xfffff)<<4);
}

void send_dac_cmd(int port, int reg, int value)
{
int mode=0;

switch(port) {
	case 1: mode=args_info.portA_mode_arg;
		break;
	case 2: mode=args_info.portB_mode_arg;
		break;
	case 3: mode=args_info.portC_mode_arg;
		break;
	}

switch(mode) {
	case 2:
		CONTROL_CMD_START(2);
		DAC8734_CMD(port, reg, value);
		NOP_CMD;
		break;
	case 4:
		CONTROL_CMD_START(2);
		DAC1220_CMD(port, reg, 0, (value & 0xff)<<16);
		NOP_CMD;
		break;
	}
flush_out_buffer();
}

void set_dac(int port, int channel, double value)
{
int mode=0;
if(channel<0 || channel > 3) {
	fprintf(stderr, "DAC has only 4 channels, 0-3\n");
	exit(-1);
	}

switch(port) {
	case 1: mode=args_info.portA_mode_arg;
		break;
	case 2: mode=args_info.portB_mode_arg;
		break;
	case 3: mode=args_info.portC_mode_arg;
		break;
	}

switch(mode) {
	case 2: 
		CONTROL_CMD_START(2);
		DAC8734_OUTPUT_CMD(port, channel, volts_to_DAC8734_code(value), 1);
		NOP_CMD
		NOP
		NOP
		break;
	case 4:
		CONTROL_CMD_START(2);
		DAC1220_CMD(port, 0, 2, volts_to_DAC1220_code(value));
		NOP_CMD
		NOP
		NOP
		break;
	}

flush_out_buffer();
}

void set_dac4x(int port, double ch0, double ch1, double ch2, double ch3)
{
CONTROL_CMD_START(5);
DAC8734_OUTPUT_CMD(port, 0, volts_to_DAC8734_code(ch0), 0);
DAC8734_OUTPUT_CMD(port, 1, volts_to_DAC8734_code(ch1), 0);
DAC8734_OUTPUT_CMD(port, 2, volts_to_DAC8734_code(ch2), 0);
DAC8734_OUTPUT_CMD(port, 3, volts_to_DAC8734_code(ch3), 1);
NOP_CMD
NOP
NOP
flush_out_buffer();
}

void set_leds(int x)
{
if(out_cmd_free+2>=MAX_BUFFER_LEN)flush_out_buffer();

SET_REGISTER(REG_LEDS, x);

flush_out_buffer();
}

void request_status(void)
{
if(out_cmd_free+2>=MAX_BUFFER_LEN)flush_out_buffer();

SET_REGISTER(REG_GENERAL_CONFIG, 1);

flush_out_buffer();
}