#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "daemon.h"
#include "control_interface.h"
#include "send_cmd.h"
#include "tiltmeter.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>


gsl_rng *rng=NULL;

static int phase=0;
static int prev_phases=-1;

static double phase_weights[8]={0, 1/6.0, 1/6.0, 1/6.0, 0, -1/6.0, -1/6.0, -1/6.0 };
//double phase_weights[8]={0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125, 0.125};

static double lvdt_accum[8]={0, 0, 0, 0, 0, 0, 0, 0};
static double level_accum[8]={0, 0, 0, 0, 0, 0, 0, 0};

static int slow_count=0;
static double lvdt_slow[8]={0, 0, 0, 0, 0, 0, 0, 0};

static int phase_count=0;
static double phase_accum[64];

static int coarse_position=1;

static double rate=50;
static double delta_t=1.0/50;

static double working_point=0.0;
static double working_point_mad=0.0;
static double injection=0.0;


static double kick_constant=0.0;
static double damping_constant=5.0;
static double damping_highpass=10;
static double damping_slew_rate=0.1;
static double integrator_slew_rate=0.0;
static double integrator_accum=0.0;
static double integrator_bang_strength=0.0;
static double damping_accum=0.0;
static double antispring=0.0;

static double sweep_start=1.0;
static double sweep_stop=1.0;
static double sweep_rate=0.01;
static double sweep_amplitude=0.1;

static double sine_frequency=1;
static double sine_phase=0;
static double sine_amplitude=0.0;

static double noise_amplitude=0.0;

static double cycle_burst_amplitude=0;
static double cycle_burst_frequency=0.01;
static double cycle_burst_max_cycles=1;

static int injection_mask=0;


#define MODE_NONE	0
#define MODE_LOCK	1
#define MODE_DAMPING_RAMP_DOWN  2
#define MODE_SINGLE_CYCLE  3
static int control_mode=MODE_NONE;

static double dac[4];

static VARIABLE vars[]={
	DECLARE_VAR(working_point, TYPE_DOUBLE),
        DECLARE_VAR(kick_constant, TYPE_DOUBLE),
        DECLARE_VAR(damping_constant, TYPE_DOUBLE),
        DECLARE_VAR(damping_highpass, TYPE_DOUBLE),
        DECLARE_VAR(damping_slew_rate, TYPE_DOUBLE),
        DECLARE_VAR(integrator_slew_rate, TYPE_DOUBLE),
        DECLARE_VAR(integrator_bang_strength, TYPE_DOUBLE),
        DECLARE_VAR(integrator_accum, TYPE_DOUBLE),
        DECLARE_VAR(damping_accum, TYPE_DOUBLE),
        DECLARE_VAR(antispring, TYPE_DOUBLE),
        DECLARE_VAR(control_mode, TYPE_INT),

	DECLARE_VAR(sweep_start, TYPE_DOUBLE),
        DECLARE_VAR(sweep_stop, TYPE_DOUBLE),
        DECLARE_VAR(sweep_rate, TYPE_DOUBLE),
        DECLARE_VAR(sweep_amplitude, TYPE_DOUBLE),
        
        DECLARE_VAR(sine_frequency, TYPE_DOUBLE),        
        DECLARE_VAR(sine_phase, TYPE_DOUBLE),        
        DECLARE_VAR(sine_amplitude, TYPE_DOUBLE),        

	DECLARE_VAR(noise_amplitude, TYPE_DOUBLE),      
        
        DECLARE_VAR(cycle_burst_frequency, TYPE_DOUBLE),        
        DECLARE_VAR(cycle_burst_amplitude, TYPE_DOUBLE),        
        DECLARE_VAR(cycle_burst_max_cycles, TYPE_DOUBLE),        

	DECLARE_VAR(injection_mask, TYPE_INT),
        
	END_VARS
	};

void init_tiltmeter(void)
{
gsl_rng_env_setup();
rng=gsl_rng_alloc(gsl_rng_default);
add_variables(vars, -1);
}

void send_header(void)
{
broadcast("LVDT_HEADER timestamp sample_count AUX1 AUX2 AUX3 AUX4 lvdt1_coarse lvdt1_fine lvdt2_coarse lvdt2_fine dac0 dac1 dac2 dac3 working_point_mad injection inj_mask level0 level1 level2 level3 level4 level5 level6 level7\n");
}

void tiltmeter_process_adc_packet(int channel, int phases, double sample_count, double *adc, int nadc)
{
int i;
struct timeval tv;
//fprintf(stderr, "phases=%d prev_phases=%d\n", phases, prev_phases);
if(phases!=0 && prev_phases==0) {
	phase=0;
	} else phase++;

prev_phases=phases;

if(0){
double x=(((unsigned long)sample_count) % 1000000)*4e-5;

if(x>20.0)x=40-x;
x-=10.0;

if((((unsigned long)sample_count) & 0x1f)==0) set_dac4x (2, x, 6.666666, x, x);
}

// fprintf(stderr, "ADC");
// for(i=0;i<nadc;i++)fprintf(stderr, " %g", adc[i]);
// fprintf(stderr, "\n");

if(phase<8) {
        for(i=0;i<nadc;i++)
                lvdt_accum[i]+=adc[i]*phase_weights[phase];
	for(i=0;i<nadc;i++)
        	level_accum[i]+=adc[i];
        } else {
        fprintf(stderr, "phase=%d %x %x\n", phase, phases, prev_phases);
        phase=0;
        }

if(phase==7) {
	for(i=0;i<8;i++)
		lvdt_slow[i]+=lvdt_accum[i];

	for(i=0;i<8;i++)
		lvdt_accum[i]=0;

	
	slow_count++;
	if(slow_count>=125) {
		gettimeofday(&tv, NULL);

		for(i=0;i<8;i++) {
			level_accum[i]/=slow_count*8.0;
			lvdt_slow[i]/=slow_count;
			}
		
		update_position(sample_count, lvdt_slow[4], lvdt_slow[5], lvdt_slow[6], lvdt_slow[7]);
		broadcast("LVDT_V %.8f %.0f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %d %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f\n", 
			(1.0*tv.tv_sec+1e-6*tv.tv_usec), sample_count, 
			lvdt_slow[0], lvdt_slow[1], lvdt_slow[2], lvdt_slow[3], lvdt_slow[4], lvdt_slow[5], lvdt_slow[6], lvdt_slow[7],
			dac[0], dac[1], dac[2], dac[3], working_point_mad, injection, injection_mask,
			level_accum[0], level_accum[1], level_accum[2], level_accum[3], level_accum[4], level_accum[5], level_accum[6], level_accum[7]
			);
		slow_count=0;
		for(i=0;i<8;i++) {
			lvdt_slow[i]=0;		
			level_accum[i]=0;
			}
		}
	}
	
if(phase_count==0) {
	for(i=0;i<8;i++)
		broadcast("PHASE %d %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f\n", i, 
			phase_accum[8*i], 
			phase_accum[8*i+1], 
			phase_accum[8*i+2], 
			phase_accum[8*i+3], 
			phase_accum[8*i+4], 
			phase_accum[8*i+5], 
			phase_accum[8*i+6], 
			phase_accum[8*i+7]);
						
			
	for(i=0;i<64;i++)phase_accum[i]=0.0;
	}

if(phase<8) {
	for(i=0;i<nadc;i++)
		phase_accum[phase+8*i]+=adc[i];
	}

phase_count++;
if(phase_count==8000)phase_count=0;

}

void mode_done(void)
{
control_mode=MODE_NONE;
broadcast_variables(vars, -1);
broadcast("MODE_SWITCH\n");
}

static int lock_phase=0;

void update_lock_mode(void)
{
switch(lock_phase) {
	case 0:
		integrator_slew_rate=10;
		integrator_bang_strength=10000;
		antispring=-1;
		damping_constant=10;
		damping_highpass=10;
		broadcast_variables(vars, -1);
		lock_phase=1;
		break;
	case 1:			
		if(working_point_mad<0.2) {
			integrator_slew_rate=10;
			integrator_bang_strength=100;
			antispring=-1;
			damping_constant=10;
			damping_highpass=10;
			broadcast_variables(vars, -1);
			lock_phase=2;
			}
		break;
	case 2: 
		if(working_point_mad<0.01) {
			integrator_slew_rate*=0.999;
			integrator_bang_strength*=0.999;
			antispring*=0.999;
			damping_constant=5;
			damping_highpass=10;
			broadcast_variables(vars, -1);
			} else {
			integrator_slew_rate=1;
			}
		if(integrator_bang_strength<0.01 && integrator_slew_rate<0.001)
			lock_phase=3;
		break;
	case 100:			
		if(working_point_mad<0.01) {
			lock_phase=0;
			integrator_slew_rate=0.001;
			integrator_bang_strength=0;
			antispring=0;
			damping_constant=5;
			damping_highpass=10;
			mode_done();
			}
		break;
	default:
		lock_phase=100;
		break;
	}
}

void update_damping_ramp_down(void)
{
damping_constant*=exp(-damping_slew_rate/rate);
broadcast("SET_VARIABLE damping_constant %.8f\n", damping_constant);
if(fabs(damping_constant)<1e-6) {
	damping_constant=0.0;
	mode_done();
	}
}

static int cycle_burst_phase=0;
static int cycle_burst_count=0;

void update_cycle_burst(void)
{
double cycles;
switch(cycle_burst_phase) {
	case 0: 
		cycle_burst_count=0;
		cycle_burst_phase=1;
		if(cycle_burst_frequency<=0)cycle_burst_frequency=1;
		kick_constant=cycle_burst_amplitude*0.5*(1-cos(2*M_PI*cycle_burst_count*cycle_burst_frequency/rate));
		broadcast_variables(vars, -1);
		break;
	case 1:
		cycle_burst_count++;
		cycles=cycle_burst_count*cycle_burst_frequency/rate;
		if(cycles>cycle_burst_max_cycles) {
			kick_constant=0;
			cycle_burst_phase=0;
			mode_done();
			return;
			}
		kick_constant=cycle_burst_amplitude*0.5*(1-cos(2*M_PI*cycles));
		broadcast("SET_VARIABLE kick_constant %.8f\n", kick_constant);
		break;
	default:
		cycle_burst_phase=0;
	}
}

void update_mode(void)
{
switch(control_mode) {
	case MODE_NONE: 
			lock_phase=0;
			cycle_burst_phase=0;
			break;
	case MODE_LOCK: update_lock_mode();
			break;
	case MODE_DAMPING_RAMP_DOWN:
			update_damping_ramp_down();
			break;
	case MODE_SINGLE_CYCLE:
			update_cycle_burst();
			break;
	}
}

void update_position(double sample_count, double ch0, double ch1, double ch2, double ch3)
{
double position;
double damping_force;
double q;
int i;

if(coarse_position && fabs(ch1)<1 && fabs(ch3)<1) {
	coarse_position=0;
	}
	
if(!coarse_position && (fabs(ch1)>2 || fabs(ch3)>2)) {
	coarse_position=1;
	}

if(coarse_position) {
	position=ch2-ch0;
//	fprintf(stderr, "c %lf\n", position);
	} else {
	position=(ch3-ch1)/5.25;
//	fprintf(stderr, "f %lf\n", position);
	}
	
if(!isfinite(position))position=100.0;

//if (position<-1.6 || position>0.55) {
//	integrator_slew_rate=-0.1; 
//	} else integrator_slew_rate=-0.01;

integrator_accum+=-((position-working_point)*integrator_slew_rate)/rate;
if (integrator_accum>10) {
	integrator_accum=10; }
if (integrator_accum<-10) {
	integrator_accum=-10; }

working_point_mad=working_point_mad*0.999+0.001*fabs(position-working_point);

//fprintf(stderr, "i   %lf\n", integrator_accum);

q=exp(-damping_highpass/rate);
damping_accum=damping_accum*q+position*(1-q);

damping_force=damping_constant*(damping_accum-position)*rate;

//fprintf(stderr, "d       %lf %lf\n", damping_accum, damping_force);
//fprintf(stderr, "integrator_slew_rate:       %lf %lf\n", integrator_slew_rate, damping_highpass);

injection=sine_amplitude*sin(2.0*M_PI*(sine_frequency*sample_count-floor(sine_frequency*sample_count))+sine_phase)+gsl_ran_gaussian(rng, noise_amplitude);

dac[0]=integrator_accum-integrator_bang_strength*(position-working_point);
dac[1]=kick_constant+damping_force+antispring*(position-working_point);
dac[2]=0.0;
dac[3]=0.0;

for(i=0;i<4;i++)dac[i]+=injection*((injection_mask>>i) & 0x1);

set_dac4x (2, dac[0], dac[1], dac[2], dac[3]);

update_mode();
}

